from dataclasses import dataclass

from faas.context.platform import FunctionReplicaService, NetworkService, TelemetryService, FunctionDeploymentService, \
    NodeService, TraceService, ZoneService


@dataclass
class PlatformContext:
    deployment_service: FunctionDeploymentService
    network_service: NetworkService
    node_service: NodeService
    replica_service: FunctionReplicaService
    telemetry_service: TelemetryService
    trace_service: TraceService
    zone_service: ZoneService



class Optimizer:
    """
    This class represents the base class optimization implementations for different resource
    management techniques.
    Each implementation must implement the run method.
    The run method starts the optimization indefinitely.
    A commonly used method is to implement a reconciliation loop that repeatedly makes decision.
    But it is also possible to implement other strategies (i.e., event-based).
    The only constraint is, that executing the `run` method will start this process.
    Initialization can take place in the implementation's constructor or in the setup method that must be called
    before executing the run method.
    The stop method should abort the process and cleanup any open resources.
    """

    def setup(self): ...

    def run(self): ...

    def stop(self): ...


from typing import List

from faas.context import ZoneService


class InMemoryZoneService(ZoneService):

    def __init__(self, zones: List[str]):
        self.zones = zones

    def get_zones(self) -> List[str]:
        return self.zones



import abc
from typing import List


class ZoneService(abc.ABC):

    def get_zones(self) -> List[str]:
        """
        Get all zones that are available in the cluster
        :return: a list of zones in the cluster
        """
        raise NotImplementedError()



import json
import os
from typing import Tuple, Dict

from faas.util.rwlock import ReadWriteLock
from .api import NetworkService


class InMemoryNetworkService(NetworkService):

    def __init__(self, min_latency: float, max_latency: float, latency_map: Dict[Tuple[str, str], float]):
        self.max_latency = max_latency
        self.min_latency = min_latency
        self.latency_map = latency_map
        self.rw_lock = ReadWriteLock()

    def get_latency(self, from_node: str, to_node: str) -> float:
        with self.rw_lock.lock.gen_rlock():
            latency = self.latency_map.get((from_node, to_node), None)
            if latency is None:
                latency = self.latency_map.get((to_node, from_node), None)
                if latency is None:
                    raise ValueError(f'No latency for connection: {from_node} - {to_node}')
            return latency

    def get_max_latency(self) -> float:
        return self.max_latency

    def update_latency(self, from_node: str, to_node: str, value: float):
        with self.rw_lock.lock.gen_wlock():
            curr_avg = self.latency_map[(from_node, to_node)]
            if curr_avg == 0:
                new_avg = value
            else:
                new_avg = (curr_avg + value) / 2
            # if new_avg > self.max_latency:
            #     self.max_latency = new_avg
            self.latency_map[(from_node, to_node)] = new_avg
            self.latency_map[(to_node, from_node)] = new_avg

    @classmethod
    def from_env(cls) -> 'InMemoryNetworkService':
        json_file = os.environ.get('edgerun_faas_latency_graph_json', None)
        max_latency = int(os.environ.get('edgerun_faas__context_max_latency', 0))
        min_latency = None
        if json_file is not None:
            with open(json_file, 'r') as fd:
                json_latency_map: Dict[str, Dict[str, float]] = json.load(fd)
                latency_map = {}
                for from_node, to_values in json_latency_map.items():
                    for to_node, latency in to_values.items():
                        latency_map[(from_node, to_node)] = latency
                        if latency > max_latency:
                            max_latency = latency
                        if min_latency is None or min_latency > latency:
                            min_latency = latency
                    latency_map[(from_node, from_node)] = 0
                return InMemoryNetworkService(min_latency, max_latency, latency_map)
        else:
            return InMemoryNetworkService(-1, -1, {})



import abc
from datetime import datetime
from typing import Dict, Callable, NamedTuple

from faas.system.core import FunctionContainer, FunctionDeployment, FunctionReplica, ScheduleEvent, FunctionReplicaState


class Record(NamedTuple):
    measurement: str
    time: int
    fields: Dict
    tags: Dict


class Clock:
    def now(self) -> datetime:
        raise NotImplementedError()


class WallClock(Clock):

    def now(self) -> datetime:
        return datetime.now()


class MetricsLogger(abc.ABC):
    def log(self, metric, value, time=None, **tags):
        """
        Call l.log('cpu_load', .65, host='server0', region='us-west') or

        :param metric: the name of the measurement
        :param value: the measurement value
        :param time: the (optional) time, otherwise now will be used
        :param tags: additional tags describing the measurement
        :return:
        """
        ...


class NullLogger(MetricsLogger):
    """
    Null logger does nothing.
    """

    def log(self, metric, value, time=None, **tags):
        pass


class RuntimeLogger(MetricsLogger):
    def __init__(self, clock=None) -> None:
        self.records = list()
        self.clock = clock or WallClock()

    def get(self, name, **tags):
        return lambda x: self.log(name, x, None, **tags)

    def log(self, metric, value, time=None, **tags):
        """
        Call l.log('cpu_load', .65, host='server0', region='us-west') or

        :param metric: the name of the measurement
        :param value: the measurement value
        :param time: the (optional) time, otherwise now will be used
        :param tags: additional tags describing the measurement
        :return:
        """
        if time is None:
            time = self._now()

        if type(value) == dict:
            fields = value
        else:
            fields = {
                'value': value
            }

        self._store_record(Record(metric, time, fields, tags))

    def _store_record(self, record: Record):
        self.records.append(record)

    def _now(self):
        return self.clock.now()


class PrintLogger(RuntimeLogger):

    def _store_record(self, record: Record):
        super()._store_record(record)
        print('[log]', record)


class LoggingLogger(RuntimeLogger):
    """
    Logger that uses a custom callable
    """

    def __init__(self, logger: Callable[[Dict], None], clock=None):
        super().__init__(clock)
        self.logger = logger

    def _store_record(self, record: Record):
        super()._store_record(record)
        self.logger(record.__dict__)


class Metrics:
    """
    Central logger for system events (i.e., scale down, deploy, remove...).
    Offers pre-defined log functions that are encouraged to use but allows to simply use the log function of the
    logger directly too to guarantee flexibility.
    """

    def __init__(self, log: MetricsLogger = None) -> None:
        super().__init__()
        self.logger: MetricsLogger = log or NullLogger()

    def log(self, metric, value, **tags):
        return self.logger.log(metric, value, **tags)

    def log_function_deployment(self, fn: FunctionDeployment):
        """
        Logs the deployment of a FunctionDeployment. Additionally logs all containers that are contained in it.
        """
        record = {'name': fn.name, 'labels': fn.fn.labels}
        for container in fn.fn_containers:
            self.log_function_containers(fn.name, container)
        self.log('function_deployments', record, type='deploy')

    def log_function_deployment_suspend(self, fn: FunctionDeployment):
        """
        Logs the suspension of a FunctionDeployment
        """
        record = {'name': fn.name, 'labels': fn.fn.labels}
        self.log('function_deployments', record, type='suspend')

    def log_function_deployment_remove(self, fn: FunctionDeployment):
        """
        Logs the removal of a FunctionDeployment
        :param fn:
        :return:
        """
        record = {'name': fn.name, 'labels': fn.fn.labels}
        self.log('function_deployments', record, type='remove')

    def log_function_containers(self, fn_name: str, fn: FunctionContainer):
        """
        Logs the functions name, container image and their metadata
        """
        record = {'name': fn_name, 'image': fn.image, 'labels': fn.labels, 'resource_config': fn.resource_config}
        self.log('functions_containers', record)

    def log_function_replica(self, replica: FunctionReplica, **kwargs):
        """
        Log function to store information about the replica (i.e., function name, image, labels).
        Does not contain any information about the state (see 'log_replica_lifecycle')
        :param replica:
        :return:
        """
        container = replica.container
        record = {'function': replica.function.name, 'labels': replica.labels, 'image': container.image}
        self.log('function_replicas', record, replica_id=replica.replica_id, **kwargs)

    def log_scaling(self, function_name: str, num_replicas: int, **kwargs):
        """
        Log function to indicate a scaling event for a given function.
        Number of replicas indicate how many replicas are started (or teared down in case the number is negative)
        :param function_name:
        :param num_replicas:
        :param kwargs:
        :return:
        """
        self.log('scale', num_replicas, function_name=function_name, **kwargs)

    def log_replica_lifecycle(self, replica: FunctionReplica, event: FunctionReplicaState, **kwargs):
        """
        Use to log if lifecycle events happen.
        Does not contain any info (i.e., container image). Use 'log_function_replica' to save any metadata.
        :param replica: the replica which lifecycle is changed
        :param event: the lifecycle event
        """
        node_name = "None"
        if replica.node is not None:
            node_name = replica.node.name
        self.log('replica_deployment', event, function_name=replica.function.name, node_name=node_name,
                 replica_id=replica.replica_id, **kwargs)

    def log_replica_schedule_event(self, replica: FunctionReplica, event: ScheduleEvent, **kwargs):
        """
        Logs events that revolve around the scheduling phase of a replica.
        :param replica: the replica that is being scheduled
        :param event: the event
        :param kwargs: passed onto the logger to allow extra key-value pairs
        """
        name = replica.fn_name
        image = replica.image
        self.log('schedule', event, function_name=name, image=image,
                 replica_id=replica.replica_id, **kwargs)



import abc
from typing import Dict, List, Optional, TypeVar, Generic

from ....system.core import FunctionNode, NodeState

I = TypeVar('I', bound=FunctionNode)


class NodeService(abc.ABC, Generic[I]):

    def get_zones(self) -> List[str]: ...

    def get_nodes(self) -> List[I]: ...

    def get_nodes_by_name(self) -> Dict[str, I]: ...

    def find(self, node_name: str) -> Optional[I]: ...

    def find_nodes_in_zone(self, zone: str) -> List[I]: ...

    def add_node(self, node: I): ...

    def remove_nod(self, node: str): ...

    def set_state(self, node: str, state: NodeState): ...



import logging
from typing import Dict, List, Optional, TypeVar, Callable, Union

from faas.util.constant import function_replica_delete, function_replica_add, function_replica_scale_up, \
    function_replica_scale_down, function_replica_state_change, function_replica_shutdown
from faas.util.rwlock import ReadWriteLock
from ...observer.api import Observer

logger = logging.getLogger(__name__)

from faas.system import FunctionReplicaState, FunctionReplica, FunctionDeployment
from faas.context.platform.node.api import NodeService
from faas.context.platform.deployment.api import FunctionDeploymentService
from .api import FunctionReplicaService, FunctionReplicaFactory

I = TypeVar('I', bound=FunctionReplica)
D = TypeVar('D', bound=FunctionDeployment)


class InMemoryFunctionReplicaService(FunctionReplicaService[I]):
    """
    A thread-safe static implementation of the FunctionReplicaService interface that stores all replicas in memory.
    """

    def __init__(self, node_service: NodeService, deployment_service: FunctionDeploymentService[D],
                 replica_factory: FunctionReplicaFactory[D, I]):
        super().__init__()
        self.node_service = node_service
        self.deployment_service = deployment_service
        self.replica_factory = replica_factory
        self._replicas: Dict[str, I] = {}
        self.rw_lock = ReadWriteLock()
        self.observers: List[Observer] = []

    def get_function_replicas(self) -> List[I]:
        with self.rw_lock.lock.gen_rlock():
            replicas = list(self._replicas.values())
            return replicas

    def find_by_predicate(self, predicate: Callable[[I], bool], running: bool = True,
                          state: FunctionReplicaState = None) -> \
            List[I]:
        found = None
        with self.rw_lock.lock.gen_rlock():
            replicas = self.get_function_replicas()
            for replica in replicas:
                if running and replica.state != FunctionReplicaState.RUNNING:
                    continue
                if state is not None and replica.state != state:
                    continue
                if predicate(replica):
                    found = replica
                    break
            return found

    def get_function_replicas_of_deployment(self, fn_deployment_name, running: bool = True,
                                            state: FunctionReplicaState = None) -> List[
        I]:
        with self.rw_lock.lock.gen_rlock():
            logger.debug(f'get_function_replicas_of_deployment for {fn_deployment_name}')
            containers = []
            for replica in self._replicas.values():
                if running and replica.state != FunctionReplicaState.RUNNING:
                    continue
                if state is not None and replica.state != state:
                    continue
                if fn_deployment_name == replica.fn_name:
                    containers.append(replica)
            logger.debug(f'found {len(containers)} containers for {fn_deployment_name}')
            return containers

    def find_function_replicas_with_labels(self, labels: Dict[str, str] = None, node_labels=None, running: bool = True,
                                           state: str = None) -> List[
        I]:
        if node_labels is None:
            node_labels = dict()
        logger.debug(f"find containers with labels: {labels}, and node labels: {node_labels}")
        pods = []
        with self.rw_lock.lock.gen_rlock():
            for replica in self._replicas.values():
                if running and replica.state != FunctionReplicaState.RUNNING:
                    continue
                if state is not None and replica.state != state:
                    continue
                matches = True
                if labels is not None:
                    pod_labels = replica.labels
                    matches = self.matches_labels(labels, pod_labels)
                node = self.node_service.find(replica.node.name)
                if not matches or node is None:
                    continue
                if node_labels is not None:
                    matches = matches and self.matches_labels(node_labels, node.labels)
                if matches:
                    pods.append(replica)
            return pods

    def get_function_replica_by_id(self, replica_id: str) -> Optional[I]:
        with self.rw_lock.lock.gen_rlock():
            replicas = list(filter(lambda r: r.replica_id == replica_id, self._replicas.values()))
            replica = None
            if len(replicas) == 1:
                replica = replicas[0]
            return replica

    def get_function_replica_with_id(self, replica_id: str) -> Optional[I]:
        logger.debug(f'find replica with id: {replica_id}')
        with self.rw_lock.lock.gen_rlock():
            found = None
            for replica in self._replicas.values():
                if replica.replica_id == replica_id:
                    found = replica
                    break
            return found

    def get_function_replicas_on_node(self, node_name: str) -> List[I]:
        logger.debug(f'find containers on ode {node_name}')
        with self.rw_lock.lock.gen_rlock():
            replicas = list(filter(lambda r: r.node.name == node_name, self._replicas.values()))
            return replicas

    def shutdown_function_replica(self, replica_id: str):
        logger.info(f"Shutdown replica with ID: {replica_id}")
        with self.rw_lock.lock.gen_wlock():
            self._shutdown_function_replica(replica_id)
        replica = self._replicas[replica_id]
        for observer in self.observers:
            payload = {
                'request': replica,
                'response': replica
            }
            observer.fire(function_replica_shutdown, payload)

    def _shutdown_function_replica(self, replica_id: str):
        try:
            self._replicas[replica_id].state = FunctionReplicaState.SHUTDOWN
        except KeyError:
            logger.info(f'Wanted to shutdown non existing replica with ID "{replica_id}"')

    def delete_function_replica(self, replica_id: str):
        logger.info(f"Shutdown replica with ID: {replica_id}")
        with self.rw_lock.lock.gen_wlock():
            self._delete_function_replica(replica_id)
            replica = self._replicas[replica_id]
        for observer in self.observers:
            payload = {
                'request': replica,
                'response': replica
            }
            observer.fire(function_replica_delete, payload)

    def _delete_function_replica(self, replica_id: str):
        try:
            self._replicas[replica_id].state = FunctionReplicaState.DELETE
        except KeyError:
            logger.info(f'Wanted to delete non existing replica with ID "{replica_id}"')

    def add_function_replica(self, replica: I):
        logger.info(f"Add replica with ID: {replica.replica_id}")
        with self.rw_lock.lock.gen_wlock():
            self._add_function_replica(replica)
        for observer in self.observers:
            payload = {
                'request': replica,
                'response': replica,
            }
            observer.fire(function_replica_add, payload)

    def _add_function_replica(self, replica: I) -> I:
        stored_replica = self._replicas.get(replica.replica_id, None)
        if stored_replica is not None:
            if stored_replica.state == FunctionReplicaState.PENDING or stored_replica.state == FunctionReplicaState.RUNNING:
                # only update pod in case it is pending or running, prevents from updating a not running to running pod
                logger.info(f"Update replica: {replica}")
                self._replicas[replica.replica_id] = replica
        else:
            logger.info(f"Create replica: {replica}")
            self._replicas[replica.replica_id] = replica
        return replica

    def matches_labels(self, labels, to_match):
        for label in labels:
            to_match_value = to_match.get(label, None)
            if to_match_value is None or to_match_value != labels[label]:
                return False
        return True

    def scale_down(self, function_name: str, remove: Union[int, List[I]]) -> List[I]:
        with self.rw_lock.lock.gen_wlock():
            if self.deployment_service.get_by_name(function_name) is None:
                raise ValueError(f'FunctionDeployment {function_name} does not exist.')
            if type(remove) is int:
                replicas = self._replicas[function_name]
                all_replicas = len(replicas)
                # just choose the last ones that were added
                removed_replicas: List[I] = replicas[all_replicas - remove:]
                for removed in removed_replicas:
                    self._delete_function_replica(removed.replica_id)
                payload = {
                    'request': remove,
                    'response': removed_replicas
                }
                for observer in self.observers:
                    observer.fire(function_replica_scale_down, payload)
                return removed_replicas
            elif type(remove) is list:
                for replica in remove:
                    self._delete_function_replica(replica.replica_id)
                payload = {
                    'request': remove,
                    'response': remove
                }
                for observer in self.observers:
                    observer.fire(function_replica_scale_down, payload)
                return remove
            else:
                raise ValueError(f'Unknown type {type(remove)} for remove argument')

    def scale_up(self, function_name: str, add: Union[int, List[I]]) -> List[I]:
        with self.rw_lock.lock.gen_wlock():
            if self.deployment_service.get_by_name(function_name) is None:
                raise ValueError(f'FunctionDeployment {function_name} does not exist.')
            if type(add) is int:
                replicas = []
                fn = self.deployment_service.get_by_name(function_name)
                for i in range(add):
                    container = fn.deployment_ranking.get_first()
                    replica = self.replica_factory.create_replica(container.labels, container, fn)
                    replicas.append(replica)
                    self._add_function_replica(replica)
                payload = {
                    'request': add,
                    'response': replicas
                }
                for observer in self.observers:
                    observer.fire(function_replica_scale_up, payload)
                return replicas
            elif type(add) is List[I]:
                for replica in add:
                    self._add_function_replica(replica)
                payload = {
                    'request': add,
                    'response': add
                }
                for observer in self.observers:
                    observer.fire(function_replica_scale_up, payload)
                return add
            else:
                raise ValueError(f'Unknown type {type(add)} for add argument')

    def register(self, observer: Observer):
        self.observers.append(observer)
        for replica in self.get_function_replicas():
            observer.fire(function_replica_add, {'request': replica, 'response': replica})

    def set_state(self, replica: FunctionReplica, state: FunctionReplicaState):
        old_state = replica.state
        with self.rw_lock.lock.gen_wlock():
            replica.state = state
        payload = {
            'replica': replica,
            'old': old_state,
            'new': state
        }
        for observer in self.observers:
            observer.fire(function_replica_state_change, payload)



import abc
from typing import Generic, TypeVar, Callable, List

import pandas as pd

from faas.system import FunctionResponse

I = TypeVar('I', bound=FunctionResponse)


class TraceService(abc.ABC, Generic[I]):

    def get_traces_api_gateway(self, node_name: str, start: float, end: float,
                               response_status: int = None) -> pd.DataFrame:
        """
        Returns all traces that were processed in the region the passed api gateway is situated.
        Which means that the API gateway was the last one to forward the request to the compute unit (i.e., Pod)

        DataFrame contains:
            'ts'
            'function'
            'image'
            'container_id'
            'node'
            'rtt'
            'sent'
            'done'
            'origin_zone'
            'dest_zone'
            'client

        :param node_name: the name of the node which hosts the gateway
        :return: DataFrame containing the traces
        """
        ...

    def get_traces_for_function(self, function: str, start: float, end: float, zone: str = None,
                                response_status: int = None):
        """
        Returns all traces that were processed for the given deployment.

        DataFrame contains:
            'ts'
            'function'
            'image'
            'function_replica_id'
            'node'
            'rtt'
            'sent'
            'done'
            'origin_zone'
            'dest_zone'
            'client

        :param function: the function to fetch functions calls from
        :param start: float
        :param end: float
        :param zone: optional, includes only traces that were processed in the given zone (i.e., the executing node was in the zone)
        :param response_status: optional, if None returns all responses otherwise checks for status code
        :return: DataFrame containing the traces
        """
        ...

    def get_values_for_function(self, function: str, start: float, end: float, access: Callable[['ResponseRepresentation'], float], zone: str = None,
                                response_status: int = None):
        """
        Returns all values from traces that were processed for the given deployment.

        :param function: the function to fetch functions calls from
        :param start: float
        :param end: float
        :param zone: optional, includes only traces that were processed in the given zone (i.e., the executing node was in the zone)
        :param response_status: optional, if None returns all responses otherwise checks for status code
        :return: DataFrame containing the traces
        """
        ...

    def get_traces_for_function_image(self, function: str, function_image: str, start: float, end: float,
                                      zone: str = None,
                                      response_status: int = None): ...

    def add_trace(self, response: I):
        ...



from .model import PlatformContext
from .platform import FunctionDeploymentService, NetworkService, NodeService, FunctionReplicaService, ZoneService, \
    TelemetryService, TraceService, InMemoryFunctionReplicaService, InMemoryNodeService, InMemoryTraceService, \
    ResponseRepresentation, InMemoryDeploymentService, FunctionReplicaFactory



from copy import copy
from typing import List, Optional, Dict, TypeVar

from faas.context.platform.node.api import NodeService
from faas.system.core import NodeState, FunctionNode
from faas.util.rwlock import ReadWriteLock

I = TypeVar('I', bound=FunctionNode)


class InMemoryNodeService(NodeService[I]):

    def __init__(self, zones: List[str], nodes: List[I]):
        self.zones = zones
        self.nodes = nodes
        self._nodes_by_name = {}
        for node in self.nodes:
            self._nodes_by_name[node.name] = node
        self.rwlock = ReadWriteLock()

    def get_zones(self) -> List[str]:
        return copy(self.zones)

    def get_nodes(self) -> List[I]:
        with self.rwlock.lock.gen_rlock():
            return copy(self.nodes)

    def get_nodes_by_name(self) -> Dict[str, I]:
        with self.rwlock.lock.gen_rlock():
            return copy(self._nodes_by_name)

    def find(self, node_name: str) -> Optional[I]:
        with self.rwlock.lock.gen_rlock():
            for node in self.nodes:
                if node.name == node_name:
                    return node
        return None

    def find_nodes_in_zone(self, zone: str) -> List[I]:
        with self.rwlock.lock.gen_rlock():
            collected = []
            for node in self.nodes:
                if node.cluster == zone:
                    collected.append(node)
            return collected

    def add_node(self, node: FunctionNode):
        with self.rwlock.lock.gen_wlock():
            self.nodes.append(node)
            self._nodes_by_name[node.name] = node

    def remove_nod(self, node: str):
        with self.rwlock.lock.gen_wlock():
            del self._nodes_by_name[node]
            self.nodes = filter(lambda n: n.name != node, self.nodes)

    def set_state(self, node: str, state: NodeState):
        with self.rwlock.lock.gen_wlock():
            self._nodes_by_name[node].state = state



from copy import copy
from typing import List, Optional, TypeVar

from faas.system import FunctionDeployment
from faas.util.rwlock import ReadWriteLock
from .api import FunctionDeploymentService

I = TypeVar('I', bound=FunctionDeployment)


class InMemoryDeploymentService(FunctionDeploymentService[I]):

    def __init__(self, deployments: List[I]):
        self.deployments = deployments
        self._deployments = {}
        self.rw_lock = ReadWriteLock()
        for deployment in deployments:
            self._deployments[deployment.name] = deployment

    def get_deployments(self) -> List[I]:
        with self.rw_lock.lock.gen_rlock():
            return copy(self.deployments)

    def get_by_name(self, fn_name: str) -> Optional[I]:
        with self.rw_lock.lock.gen_rlock():
            return self._deployments.get(fn_name)

    def exists(self, fn_name: str) -> bool:
        with self.rw_lock.lock.gen_rlock():
            return self._exists(fn_name)

    def _exists(self, fn_name: str) -> bool:
        return self._deployments.get(fn_name, None) is not None

    def remove(self, function_name: str):
        with self.rw_lock.lock.gen_wlock():
            if self._exists(function_name):
                self.deployments = [x for x in self.deployments if x.name != function_name]
                del self._deployments[function_name]

    def add(self, deployment: I):
        with self.rw_lock.lock.gen_wlock():
            if not self._exists(deployment.name):
                self.deployments.append(deployment)
                self._deployments[deployment.name] = deployment



from faas.opt.api import Optimizer


class ReconciliationOptimizationDaemon(Optimizer):

    def __init__(self, optimizer: Optimizer):
        self.is_running = True
        self.optimizer = optimizer

    def sleep(self): ...

    def setup(self):
        self.optimizer.setup()

    def run(self):
        self.optimizer.run()
        while self.is_running:
            self.sleep()
            self.optimizer.run()

    def stop(self):
        self.optimizer.stop()
        self.is_running = False




import abc
from typing import List, Optional, Dict, Callable, Generic, TypeVar, Union

from ...observer.api import Observer
from ....system.core import FunctionReplica, FunctionReplicaState, FunctionContainer, FunctionDeployment

I = TypeVar('I', bound=FunctionReplica)
D = TypeVar('D', bound=FunctionDeployment)


class FunctionReplicaService(abc.ABC, Generic[I]):

    def find_by_predicate(self, predicate: Callable[[I], bool], running: bool = True,
                          state: FunctionReplicaState = None) -> \
            List[I]:
        ...

    def get_function_replicas(self) -> List[I]:
        raise NotImplementedError()

    def get_function_replicas_of_deployment(self, deployment_name, running: bool = True,
                                            state: FunctionReplicaState = None) -> \
            List[I]:
        """
        param name: original name of the deployment
        """
        raise NotImplementedError()

    def find_function_replicas_with_labels(self, labels: Dict[str, str] = None, node_labels=None,
                                           running: bool = True, state: FunctionReplicaState = None) -> List[
        I]:
        raise NotImplementedError()

    def get_function_replica_by_id(self, replica_id: str) -> Optional[I]:
        raise NotImplementedError()

    def get_function_replicas_on_node(self, node_name: str) -> List[I]:
        raise NotImplementedError()

    def shutdown_function_replica(self, replica_id: str):
        raise NotImplementedError()

    def add_function_replica(self, functionReplica: I) -> I:
        raise NotImplementedError()

    def delete_function_replica(self, replica_id: str): ...

    def scale_down(self, function_name: str, remove: Union[int, List[I]]) -> List[I]: ...

    def scale_up(self, function_name: str, add: Union[int, List[I]]) -> List[I]: ...

    def register(self, observer: Observer): ...

    def set_state(self, replica: FunctionReplica, state: FunctionReplicaState): ...


class FunctionReplicaFactory(abc.ABC, Generic[D, I]):

    def create_replica(self, labels: Dict[str, str], fn_container: FunctionContainer,
                       fn_deployment: D) -> I: ...



from typing import TypeVar, Generic, List

T = TypeVar('T')


class Point(Generic[T]):
    ts: float
    val: T

    def __init__(self, ts: float, val: T):
        self.ts = ts
        self.val = val

    def get_ts(self) -> float:
        return self.ts

    def get_value(self) -> T:
        return self.val

    def __repr__(self):
        return f"Window(ts={self.ts}, val={self.val})"


S = TypeVar('S')


class PointWindow(Generic[S]):
    _content: List[Point[S]]

    def __init__(self, limit: int):
        """

        :param limit: specifies the time window in seconds of content stored. I.e, if limit == 15, the list contains the
        last 15 seconds
        """
        self._content = []
        self.limit = limit

    def append(self, resource_window: Point[S]):
        self._content.append(resource_window)
        # not needed because old requests are filtered through dataframe
        # self._content.sort(key=lambda x: x.ts, reverse=False)
        # diff = resource_window.ts - self.limit
        # first_ts = self._content[0].ts
        # if first_ts <= diff:
        #     self._cleanup(diff)

    def purge(self, stop: float):
        index = 0
        for i, window in enumerate(self._content):
            if window.ts >= stop:
                index = i
                break
        purged = self._content[:index]
        self._content = self._content[index + 1:]
        return purged


    def size(self) -> int:
        return len(self._content)

    def value(self) -> List[Point[S]]:
        return list(self._content)



import logging
from collections import defaultdict
from dataclasses import dataclass
from typing import Dict, TypeVar, Callable, Optional

import pandas as pd

from faas.context.platform.node.api import NodeService
from faas.context.platform.trace.api import TraceService
from faas.system import FunctionResponse
from faas.util.point import PointWindow, Point
from faas.util.rwlock import ReadWriteLock

logger = logging.getLogger(__name__)

I = TypeVar('I', bound=FunctionResponse)


@dataclass
class ResponseRepresentation:
    request_id: int
    ts: float
    function: str
    function_image: str
    replica_id: str
    node: str
    rtt: float
    done: float
    sent: float
    origin_zone: str
    dest_zone: str
    client: str
    status: int


class InMemoryTraceService(TraceService[I]):

    def __init__(self, now: Callable[[],float], window_size: int, node_service: NodeService,
                 parser: Callable[[I], Optional[ResponseRepresentation]]):
        self.now = now
        self.window_size = window_size
        self.node_service = node_service
        self.parser = parser
        self.requests_per_node: Dict[str, PointWindow[I]] = {}
        self.locks = {}
        self.last_purge = 0
        # TODO does not support new nodes during experiments
        for node in node_service.get_nodes():
            self.locks[node.name] = ReadWriteLock()

    def _purge(self, till_ts: float):
        for node, point_window in self.requests_per_node.items():
            with self.locks[node].lock.gen_wlock():
                point_window.purge(till_ts)
    def purge(self):
        now = self.now()
        duration_since_last_purge = now - self.last_purge
        if duration_since_last_purge > self.window_size:
            self._purge(now - self.window_size)
            self.last_purge = now

    def get_traces_api_gateway(self, node_name: str, start: float, end: float,
                               response_status: int = None) -> pd.DataFrame:
        self.purge()
        gateway = self.node_service.find(node_name)
        if gateway is None:
            nodes = self.node_service.get_nodes_by_name()
            raise ValueError(f"Node {node_name} not found, currently stored: {nodes}")
        zone = gateway.cluster
        nodes = self.node_service.find_nodes_in_zone(zone)
        requests = defaultdict(list)
        if len(nodes) == 0:
            logger.info(f'No nodes found in zone {zone}')
        for node in nodes:
            with self.locks[node_name].lock.gen_rlock():
                node_requests = self.requests_per_node.get(node.name)
                if node_requests is None or node_requests.size() == 0:
                    continue

                for req in node_requests.value():
                        parsed = self.parser(req.val)
                        if parsed is not None:
                            for key, value in parsed.__dict__.items():
                                requests[key].append(value)

        df = pd.DataFrame(data=requests).sort_values(by='ts')
        df.index = pd.DatetimeIndex(pd.to_datetime(df['ts'], unit='s'))

        df = df[df['ts'] >= start]
        df = df[df['ts'] <= end]
        logger.info(f'After filtering {len(df)} traces left for api gateway {node_name}')
        if response_status is not None:
            df = df[df['status'] == response_status]
            logger.info(f'After filtering out non status: {len(df)}')
        return df

    def add_trace(self, response: I):
        with self.locks[response.node.name].lock.gen_wlock():
            node = response.node.name
            window = self.requests_per_node.get(node, None)
            if window is None:
                self.requests_per_node[node] = PointWindow(self.window_size)
            self.requests_per_node[node].append(Point(response.request.start, response))

    def get_traces_for_function(self, function_name: str, start: float, end: float, zone: str = None,
                                response_status: int = None):
        self.purge()

        if zone is not None:
            nodes = self.node_service.find_nodes_in_zone(zone)
        else:
            nodes = self.node_service.get_nodes()
        requests = defaultdict(list)
        for node in nodes:
            node_name = node.name
            with self.locks[node_name].lock.gen_rlock():
                node_requests = self.requests_per_node.get(node_name)
                if node_requests is None or node_requests.size() == 0:
                    continue

                for req in node_requests.value():
                    if req.val.name == function_name:
                        parsed = self.parser(req.val)
                        if parsed is not None:
                            for key, value in parsed.__dict__.items():
                                requests[key].append(value)

        df = pd.DataFrame(data=requests)
        if len(df) == 0:
            return df
        df = df.sort_values(by='ts')
        df.index = pd.DatetimeIndex(pd.to_datetime(df['ts'], unit='s'))

        logger.info(f'Before filtering {len(df)} traces for function {function_name}')
        df = df[df['ts'] >= start]
        df = df[df['ts'] <= end]
        logger.info(f'After filtering {len(df)} traces left for function {function_name}')
        if response_status is not None:
            df = df[df['status'] == response_status]
            logger.info(f'AFter filtering out non status: {len(df)}')
        return df.reset_index(drop=True)

    def get_traces_for_function_image(self, function: str, function_image: str, start: float, end: float,
                                      zone: str = None,
                                      response_status: int = None):
        self.purge()
        df = self.get_traces_for_function(function, start, end, zone, response_status)
        df = df[df['function_image'] == function_image]
        return df

    def get_values_for_function(self, function: str, start: float, end: float,
                                access: Callable[['ResponseRepresentation'], float], zone: str = None,
                                response_status: int = None):
        if zone is not None:
            nodes = self.node_service.find_nodes_in_zone(zone)
        else:
            nodes = self.node_service.get_nodes()
        data = []
        for node in nodes:
            if zone is not None:
                if node.cluster != zone:
                    continue
            node_name = node.name
            with self.locks[node_name].lock.gen_rlock():
                node_requests = self.requests_per_node.get(node_name)
                if node_requests is None or node_requests.size() == 0:
                    continue

                for req in node_requests.value():
                    if req.val.name == function:
                        parsed = self.parser(req.val)
                        if parsed is not None:
                            if response_status is None or parsed.status == response_status:
                                if parsed.ts >= start and parsed.ts <= end:
                                    data.append(access(parsed))

        return data





import abc
from typing import List, Optional, TypeVar, Generic

from faas.system import FunctionDeployment

I = TypeVar('I', bound=FunctionDeployment)


class FunctionDeploymentService(abc.ABC, Generic[I]):

    def get_by_name(self, fn_name) -> Optional[I]:
        """
        Tries to find a FunctionDeployment which represents a Function with the given name
        :param fn_name: the function name to look for
        :return: None if no FunctionDeployment exists, otherwise a FunctionDeployment
        """
        ...

    def get_deployments(self) -> List[I]:
        """
        Returns all deployed and suspended FunctionDeployments
        :return: a list of FunctionDeployments
        """
        ...

    def exists(self, name: str) -> bool:
        """
        Checks whether a FunctionDeployment with the given name exists
        :param name: name of the FunctionDeployment
        :return: true if already deployed, false otherwise
        """
        ...

    def add(self, deployment: I):
        ...

    def remove(self, function_name: str): ...



import abc
import enum
from dataclasses import dataclass
from typing import List, Dict, Union, Any
from typing import Optional


class WorkloadConfiguration(abc.ABC):
    pass


def counter(start: int = 1):
    n = start
    while True:
        yield n
        n += 1


@dataclass
class ScalingConfiguration:
    scale_min: int = 1
    scale_max: int = 20
    # determines the percentage how many pods
    scale_factor: int = 1
    scale_zero: bool = False


class NodeState(enum.Enum):
    READY = 1
    NOT_READY = 2


class ScheduleEvent(enum.Enum):
    QUEUE = 1
    START = 2
    FINISH = 3


class FunctionReplicaState(enum.Enum):
    CONCEIVED = 1
    PENDING = 2
    RUNNING = 3
    SHUTDOWN = 4
    DELETE = 5


@dataclass
class FunctionImage:
    # the manifest list (docker image) name
    image: str

    def __init__(self, image: str):
        self.image = image


class ResourceConfiguration():

    def __init__(self, requirements: Dict[str, Any], limits: Dict[str, Any] = None):
        self.requirements = requirements
        self.limits = limits

    def get_resource_requirements(self) -> Dict:
        """
        Defines the resource requirements that a node must meet in order to eligible for scheduling.
        Has to be always defined and a default value must be provided.
        :return: resource requirements for scheduling
        """
        return self.requirements

    def get_resource_limits(self) -> Optional[Dict]:
        """
        Returns the upper limit of resource usage. Is optional and therefore can be None
        :return: an optional dict that contains upper limits for resource usage
        """
        return self.limits

    def __str__(self):
        return str(self.get_resource_requirements())

    def __repr__(self):
        return str(self.get_resource_requirements())


@dataclass
class Function:
    name: str
    fn_images: List[FunctionImage]
    labels: Dict[str, str]

    def __init__(self, name: str, fn_images: List[FunctionImage], labels: Dict[str, str] = None):
        self.fn_images = fn_images
        self.name = name
        self.labels = labels if labels is not None else {}

    def get_image(self, image: str) -> Optional[FunctionImage]:
        for fn_image in self.fn_images:
            if fn_image.image == image:
                return fn_image
        return None


@dataclass
class FunctionContainer:
    fn_image: FunctionImage
    resource_config: ResourceConfiguration
    labels: Dict[str, str]

    def __init__(self, fn_image: FunctionImage, resource_config: ResourceConfiguration,
                 labels: Dict[str, str] = None):
        self.fn_image = fn_image
        self.resource_config = resource_config
        self.labels = labels if labels is not None else {}

    @property
    def image(self):
        return self.fn_image.image

    def get_resource_requirements(self):
        return self.resource_config.get_resource_requirements()


class FunctionNode:
    name: str
    arch: str
    cpus: int
    ram: int
    netspeed: int
    labels: Dict[str, str]
    allocatable: Dict[str, str]
    cluster: Optional[str]
    state: NodeState

    def __init__(self, name: str, arch: str, cpus: int, ram: int, netspeed: int, labels: Dict[str, str],
                 allocatable: Dict[str, str], cluster: Optional[str], state: NodeState):
        self.name = name
        self.arch = arch
        self.cpus = cpus
        self.ram = ram
        self.netspeed = netspeed
        self.labels: Dict[str, str] = labels
        self.allocatable = allocatable
        self.cluster = cluster
        self.state = state


@dataclass
class DeploymentRanking:
    """
    The DeploymentRanking is used to determine at any time which DeploymentRanking to deploy for the associated
    FunctionDeployment.
    Users can modify this ranking during runtime to dynamically choose a concrete implementation (FunctionContainer)
    for a Function (FunctionDeployment).
    """
    containers: List[FunctionContainer]
    function_factor: Dict[str, float]

    def __init__(self, containers: List[FunctionContainer], function_factor: Dict[str, float] = None):
        self.containers = containers.copy()
        self.function_factor = function_factor
        if self.function_factor is None:
            self.function_factor = {}
            for container in self.containers:
                self.function_factor[container.image] = 1

    def set_first(self, container: FunctionContainer):
        """
        Sets the given container into the first position and pushes all other containers back by one.
        """
        index = self._find_index(container)
        if index is None:
            raise ValueError(f'Container {container.image} not found in ranking.')

        updated = self.containers[:index] + self.containers[index + 1:]
        self.containers = [container] + updated

    def _find_index(self, container: FunctionContainer) -> Optional[int]:
        try:
            return self.containers.index(container)
        except ValueError:
            return None

    def get_first(self):
        return self.containers[0]


@dataclass
class FunctionDeployment:
    fn: Function
    fn_containers: List[FunctionContainer]
    scaling_configuration: ScalingConfiguration
    deployment_ranking: DeploymentRanking

    def get_services(self):
        return list(map(lambda fn_container: fn_container.fn_image, self.fn_containers))

    def get_containers(self):
        return self.fn_containers

    def get_container(self, image: str) -> Optional[FunctionContainer]:
        for fn_image in self.fn_containers:
            if fn_image.image == image:
                return fn_image
        return None

    @property
    def name(self):
        return self.fn.name

    @property
    def labels(self):
        return self.fn.labels.copy()


class FunctionReplica:
    """
    A function replica is an instance of a function running on a specific node.
    """
    # name must be unique to this specific replica, i.e., UUID
    replica_id: str
    _labels: Dict[str, str]
    function: FunctionDeployment
    container: FunctionContainer
    node: Optional[FunctionNode]
    state: FunctionReplicaState

    def __init__(self, replica_id: str, labels: Dict[str, str], function: FunctionDeployment,
                 container: FunctionContainer, node: Optional[FunctionNode], state: FunctionReplicaState):
        self.replica_id = replica_id
        self._labels = labels
        self.function = function
        self.container = container
        self.node = node
        self.state = state

    @property
    def fn_name(self):
        return self.function.name

    @property
    def image(self):
        return self.container.image

    @property
    def labels(self):
        # own labels have highest priority (i.e., _labels will overwrite (function.labels | container.labels)
        labels = self.function.labels.copy()
        labels.update(self.container.labels)
        labels.update(self._labels)
        return labels


# TODO add functionrequest to trace when started

@dataclass
class FunctionRequest:
    request_id: int
    client: str
    name: str
    body: str
    start: float
    # in bytes
    size: int = None
    replica: FunctionReplica = None
    headers: Dict = None

    id_generator = counter()

    def __init__(self, name, start: float, size=None, request_id=None, body=None, client=None, replica=None,
                 headers=None) -> None:
        super().__init__()
        self.start = start
        self.name = name
        self.body = body
        self.client = client
        self.size = size
        self.request_id = request_id if request_id is not None else next(self.id_generator)
        self.replica = replica
        self.headers = headers

    def __str__(self) -> str:
        return 'FunctionRequest(%d, %s, %s)' % (self.request_id, self.name, self.size)

    def __repr__(self):
        return self.__str__()

    def __hash__(self) -> int:
        return hash(self.start) + hash(self.name) + hash(self.request_id)


@dataclass
class FunctionResponse:
    request: FunctionRequest
    request_id: Union[int, str]
    # client that has called
    client: str
    # function name
    name: str
    # response body
    body: str
    # size of the response
    size: int
    # response status code
    code: int
    # timestamp that is taken before any data transfer or execution happened
    # (ts_end - ts_start) -> is the true round trip time of the request, including every latency
    ts_start: float
    # timestamp of waiting to be executed
    ts_wait: float
    # timestamp of starting execution
    ts_exec: float
    # timestamp of ending execution
    ts_end: float
    # raw function execution time, without wait
    fet: float
    # replica which the request was processed
    replica: FunctionReplica
    # node on which the request was processed
    node: FunctionNode




class FaasSystem(abc.ABC):

    @abc.abstractmethod
    def deploy(self, fn: FunctionDeployment):
        """
        Deploys a function in the system.
        It is necessary to deploy a function before invoking, remove, etc. it.
        :param fn: function to deploy
        """
        ...

    @abc.abstractmethod
    def invoke(self, request: Union[FunctionRequest, WorkloadConfiguration]):
        """
        Invokes a function. It depends on the specific type of the argument and the implementation on how this is achieved.
        In general, passing a FunctionRequest will lead to a single invocation while a WorkloadConfiguration will specify
        a lengthy sequence of function invocations.
        :param request: either a single FunctionRequest or a specification as type of WorkloadConfiguration
        """
        ...

    @abc.abstractmethod
    def remove(self, fn: FunctionDeployment):
        """
        Removes the given function.
        This entails shutting down all function replicas. The implementation determines whether running calls are cancelled
        or will wait for all invocations being processed.
        :param fn: function to remove
        """
        ...

    @abc.abstractmethod
    def get_deployments(self) -> List[FunctionDeployment]:
        """
        Fetches all FunctionDeployments that were previously deployed. Does not contain removed deployments
        :return: a list of deployments
        """
        ...

    @abc.abstractmethod
    def get_function_index(self) -> Dict[str, FunctionContainer]:
        """
        Dictionary that contains all deployed FunctionContainer, accessible via the specific image name
        :return: a dict
        """
        ...

    @abc.abstractmethod
    def get_replicas(self, fn_name: str, running: bool = True, state=None) -> List[FunctionReplica]:
        """
        Finds all FunctionReplicas for the given function that are in the given state.
        If the state is None all replica are returned
        :param fn_name: the function name
        :param running: set to true if only RUNNING replicas should be returned or all that are not RUNNING
        :param state: optional, the state the function replicas are in or all if None
        """
        ...

    @abc.abstractmethod
    def scale_down(self, function_name: str, remove: Union[int, List[FunctionReplica]]) -> List[FunctionReplica]:
        """
        Scales down the specified function.
        Depending on argument of remove, either chooses the last 'n' replica that were started or
        shuts down the specific replicas passed.
        :param function_name:  the function to scale down
        :param remove: either a number of FunctionReplica to shut down or a specific list of replicas
        :return: list containing the removed replica
        """
        ...

    @abc.abstractmethod
    def scale_up(self, function_name: str, replicas: Union[int, List[FunctionReplica]]) -> List[FunctionReplica]:
        """
        Scales up the specified function.
        Either creates the given number of replica or uses the list of passed replicas (i.e., the node is None in each replica,
        and the implementation assigns it to one, which can happen asynchronously.
        :param function_name:  the function to scale up
        :param remove: either a number of FunctionReplica to start or a specific list of replicas (which
                       need to be mapped to specific nodes)
        :return: list containing the added replica
        """
        ...

    @abc.abstractmethod
    def discover(self, function: FunctionContainer, running: bool = True, state: FunctionReplicaState = None) -> List[
        FunctionReplica]:
        """
        Finds for a specific FunctionContainer all replicas. I.e., all replica that have the same image as the given
        FunctionContainer.
        If the state is None, returns all replicas.
        :param function:  container to look for FunctionReplica in the given state
        :param running: set to true if only RUNNING replicas should be returned or all that are not RUNNING
        :param state: optional, if replicas should be in a specific state
        :return: a list of replicas in the given state, or if None all
        """
        ...

    @abc.abstractmethod
    def poll_available_replica(self, fn: str, interval: int = 0.5, timeout: int = None) -> Optional[
        List[FunctionReplica]]:
        """
        Blocks and repeatedly checks if running replicas are available for the given function.
        :param fn: the function to poll for
        :param interval: determines in which interval the check should be made
        :param timeout: time to wait, if None waits indefinitely
        :return: None if timeout and no replicas, otherwise the running replicas
        """
        ...



from typing import List, Any

from faas.context import PlatformContext
from faas.context.observer.api import Observer
from faas.system import FunctionDeployment, FunctionReplica, FunctionReplicaState
from faas.util.constant import function_label, api_gateway_type_label, zone_label, function_replica_add, \
    function_replica_scale_up, function_replica_scale_down, function_replica_state_change, \
    function_replica_shutdown


class LoadBalancerOptimizer:
    def update(self): ...

    def get_running_replicas(self, function: str) -> List[FunctionReplica]: ...

    def get_functions(self) -> List[FunctionDeployment]: ...

    def add_replica(self, replica: FunctionReplica): ...

    def remove_replica(self, replica: FunctionReplica): ...

    def add_replicas(self, replicas: List[FunctionReplica]): ...

    def remove_replicas(self, replicas: List[FunctionReplica]): ...


class GlobalLoadBalancerOptimizer(LoadBalancerOptimizer):

    def __init__(self, context: PlatformContext):
        self.context = context
        self.observer = LoadBalancerObserver(self)
        context.replica_service.register(self.observer)

    def get_running_replicas(self, function: str) -> List[FunctionReplica]:
        replica_service = self.context.replica_service
        return replica_service.get_function_replicas_of_deployment(function, running=True)

    def get_functions(self) -> List[FunctionDeployment]:
        deployment_service = self.context.deployment_service
        functions = [d for d in deployment_service.get_deployments() if
                     d.labels.get(function_label) and d.labels.get(function_label) != api_gateway_type_label]
        return functions


class LoadBalancerObserver(Observer):

    def __init__(self, lb: LoadBalancerOptimizer):
        self.lb = lb

    def fire(self, event: str, value: Any):
        if event == function_replica_add:
            replica: FunctionReplica = value['response']
            if replica.state == FunctionReplicaState.RUNNING:
                self.lb.add_replica(replica)
        if event == function_replica_shutdown:
            replica = value['response']
            self.lb.remove_replica(replica)
        if event == function_replica_scale_up:
            replicas: List[FunctionReplica] = value['response']
            replicas = [r for r in replicas  if r.state == FunctionReplicaState.RUNNING]
            self.lb.add_replicas(replicas)
        if event == function_replica_scale_down:
            replicas = value['response']
            self.lb.remove_replicas(replicas)
        if event == function_replica_state_change:
            state = value['new']
            if state == FunctionReplicaState.RUNNING:
                self.lb.add_replica(value['replica'])

class LocalizedLoadBalancerOptimizer(LoadBalancerOptimizer):

    def __init__(self, context: PlatformContext, cluster: str):
        self.context = context
        self.cluster = cluster
        self.observer = LoadBalancerObserver(self)

    def get_functions(self) -> List[FunctionDeployment]:
        deployment_service = self.context.deployment_service
        functions = [d for d in deployment_service.get_deployments() if
                     d.labels.get(function_label) and d.labels.get(function_label) != api_gateway_type_label]
        return functions

    def get_running_replicas(self, function: str) -> List[FunctionReplica]:
        replicas = self.context.replica_service.find_function_replicas_with_labels(labels={
            function_label: function}, node_labels={zone_label: self.cluster}, running=True)

        all_load_balancers = self.context.replica_service.find_function_replicas_with_labels(labels={
            function_label: api_gateway_type_label
        })
        other_load_balancers = [l for l in all_load_balancers if l.labels[zone_label] != self.cluster]
        for lb in other_load_balancers:
            other_cluster = lb.labels[zone_label]
            other_replicas = self.context.replica_service.find_function_replicas_with_labels(labels={
                function_label: function,
            }, node_labels={zone_label: other_cluster}, running=True)
            if len(other_replicas) > 0:
                replicas.append(lb)
        return replicas



from .deployment import FunctionDeploymentService, InMemoryDeploymentService
from .network import NetworkService
from .node import NodeService, InMemoryNodeService
from .replica import FunctionReplicaService, InMemoryFunctionReplicaService, FunctionReplicaFactory
from .telemetry import TelemetryService
from .trace import TraceService, InMemoryTraceService, ResponseRepresentation
from .zone import ZoneService



zone_label = "ether.edgerun.io/zone"
pod_type_label = "ether.edgerun.io/type"
function_label = "ether.edgerun.io/function"
api_gateway_type_label = 'api-gateway'
function_type_label = 'fn'
client_role_label = 'node-role.kubernetes.io/client'
worker_role_label = 'node-role.kubernetes.io/worker'
controller_role_label = 'node-role.kubernetes.io/controller'
hostname_label = 'kubernetes.io/hostname'

# Pod status constants
pod_not_running = 'Not Running'
pod_running = 'Running'
pod_unknown = "Unknown"
pod_failed = "Failed"
pod_succeeded = "Succeeded"
pod_pending = "Pending"

# Observer events
function_replica_add = 'function-replica-add'
function_replica_delete = 'function-replica-delete'
function_replica_shutdown = 'function-replica-shutdown'
function_replica_scale_up = 'function-replica-scale-up'
function_replica_scale_down = 'function-replica-scale-down'
function_replica_state_change = 'function-replica-state-change'



import readerwriterlock.rwlock


class ReadWriteLock:
    """ A lock object that allows many simultaneous "read locks", but
    only one "write lock." """

    def __init__(self):
        self.lock = readerwriterlock.rwlock.RWLockRead()



import abc


class NetworkService(abc.ABC):

    def get_latency(self, from_node: str, to_node: str) -> float:
        raise NotImplementedError()

    def get_max_latency(self) -> float:
        raise NotImplementedError()

    def update_latency(self, from_node: str, to_noe: str, value: float):
        raise NotImplementedError()



import abc
import datetime
from typing import Optional

import pandas as pd


class TelemetryService(abc.ABC):

    def get_replica_cpu(self, fn_replica_id: str, start: float = None, end: float = None) -> \
            Optional[pd.DataFrame]:
        """
        Fetch the measured cpu times for the given replica in the given timeframe.
        If start is None, the start of the selection will be set to 01-01-1970.
        If end is None, the end will be set to now.
        """
        raise NotImplementedError()

    def get_node_cpu(self, node: str, start: float = None, end: float = None) -> Optional[
        pd.DataFrame]:
        raise NotImplementedError()
